#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""File readers.
"""

import json
import yaml
from .config import Configuration
from .regmap import BitField, Register, RegisterMap
from . import utils


class _DictReader():
    """Base class that converts file to a dictionary."""
    def _open_file(self, path):
        """Read file to dictionary."""
        with open(path, 'r') as f:
            print("  Open file ... ", end='')
            ext = utils.get_file_ext(path)
            if ext in ['.yaml', '.yml']:
                data = yaml.safe_load(f)
            elif ext == '.json':
                data = json.load(f)
            else:
                raise ValueError("Unknown extension '%s' of the file '%s'" % (ext, path))
            print("OK")
        return data


class RegisterMapReader(_DictReader):
    """Read register map and configuration from file.

    Examples:

        Read provided file and create :class:`RegisterMap` object:

        >>> from corsair import RegisterMapReader
        >>> reader = RegisterMapReader()
        >>> rmap = reader('tests/data/map.json')
        Read 'tests/data/map.json' file with RegisterMapReader:
          Open file ... OK
          Read configuration ... OK
          Read registers ... OK
    """
    def __call__(self, path, config=Configuration()):
        """Read input file.

        Returns:
            :class:`RegisterMap` object.
        """
        print("Read '%s' file with RegisterMapReader:" % path)
        data = self._open_file(path)

        # Read configuration
        print("  Read configuration ... ", end='')
        if 'config' in data.keys():
            config.values = data['config']
            print("OK")
        else:
            print("Not provided")

        # Read registers
        print("  Read registers ... ", end='')
        rmap = RegisterMap(config=config)
        for data_reg in data['regmap']:
            data_reg_filtered = {k: v for k, v in data_reg.items() if k in ['name', 'description', 'address']}
            reg = Register(**data_reg_filtered)
            for data_bf in data_reg['bfields']:
                reg.add_bfields(BitField(**data_bf))
            rmap.add_regs(reg)
        print("OK")
        return rmap


class ConfigurationReader(_DictReader):
    """Read configuration from file.

    Examples:
        Read provided file and create :class:`Configuration` object:

        >>> from corsair import ConfigurationReader
        >>> reader = ConfigurationReader()
        >>> rmap = reader('tests/data/config.json')
        Read 'tests/data/config.json' file with ConfigurationReader:
          Open file ... OK
          Read configuration ... OK
    """
    def __call__(self, path):
        """Read input file.

        Returns:
            :class:`Configuration` object.
        """
        print("Read '%s' file with ConfigurationReader:" % path)
        data = self._open_file(path)
        print("  Read configuration ... ", end='')
        config = Configuration()
        config.values = data
        print("OK")
        return config



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Utility functions and classes.
"""

import os


def hex_to_dec(val):
    """Converts hexademical string to integer"""
    return int(val, 16)


def try_hex_to_dec(val, prefix='0x'):
    """Tries to convert hexademical string to integer. Returns original value if not possible."""
    try:
        # convert only if prefixes match
        if prefix in val[:len(prefix)]:
            return hex_to_dec(val)
    except (TypeError, ValueError):
        pass
    return val


def tree_hex_to_dec(tree):
    """Recursevely convert only '0x' hexademical strings to integers inside list or dict in-place."""
    for idx, val in enumerate(tree) if isinstance(tree, list) else tree.items():
        if isinstance(val, (dict, list)):
            tree_hex_to_dec(val)
        else:
            tree[idx] = try_hex_to_dec(val)


def is_non_neg_int(val, err_msg=""):
    if isinstance(val, int) and val >= 0:
        return True
    else:
        raise ValueError("Value of '%s' must be a non-negative integer." % val if not err_msg else err_msg)


def is_pos_int(val, err_msg=""):
    if isinstance(val, int) and val > 0:
        return True
    else:
        raise ValueError("Value of '%s' must be a positive integer." % val if not err_msg else err_msg)


def listify(obj):
    """Make lists from single objects. No changes are made for the argument of the 'list' type."""
    if type(obj) is not list:
        return [obj]
    else:
        return obj


def try_int_to_str(val, max_dec=1024):
    """Make string from int. Hexademical representaion will be used if input value greater that 'max_dec'."""
    if isinstance(val, int):
        if val > max_dec:
            return "0x%x" % val
        else:
            return "%d" % val
    else:
        return val


def get_file_ext(path):
    _, ext = os.path.splitext(path)
    return ext.lower()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Corsair is a control and status register (CSR) map generator for FPGA/ASIC projects.
It generates HDL code, documentation and other artifacts from CSR map description file.
"""

__title__ = "corsair"
__description__ = "Control and status register (CSR) map generator for FPGA/ASIC projects."

try:
    from ._version import version as __version__
except (ImportError, ModuleNotFoundError) as e:
    __version__ = 'git-latest'

from .config import (
    Parameter,
    ParameterGroup,
    Configuration
)

from .regmap import (
    BitField,
    Register,
    RegisterMap
)

from .readers import (
    RegisterMapReader,
    ConfigurationReader
)

from .writers import (
    RegisterMapWriter,
    ConfigurationWriter,
    LbBridgeWriter,
    HdlWriter,
    DocsWriter,
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""File writers.
"""

import os
import json
import yaml
import jinja2
from corsair import __version__
from . import utils
from pathlib import Path
import wavedrom


class _DictWriter():
    """Base class that converts dictionary to file."""
    def _save_file(self, path, data):
        """Create file from dictionary."""
        with open(path, 'w') as f:
            print("  Save data to file ... ", end='')
            ext = utils.get_file_ext(path)
            if ext in ['.yaml', '.yml']:
                yaml.Dumper.ignore_aliases = lambda *args: True  # hack to disable aliases
                yaml.dump(data, f, default_flow_style=False, sort_keys=False)
            elif ext == '.json':
                json.dump(data, f, indent=4)
            else:
                raise ValueError("Unknown extension '%s' of the file '%s'" % (ext, path))
            print("OK")
        return data


class RegisterMapWriter(_DictWriter):
    """Write register map to a file.

    Examples:

        Create JSON file based on a :class:`RegisterMap` object:

        >>> from corsair import RegisterMapWriter, RegisterMap
        >>> writer = RegisterMapWriter()
        >>> rmap = RegisterMap()
        >>> writer('/tmp/map.json', rmap)
        Write '/tmp/map.json' file with RegisterMapWriter:
          Prepare data ... OK
          Save data to file ... OK
    """
    def __call__(self, path, rmap):
        """Write output file.

        Args:
            path : path to file
            rmap : :class:`RegisterMap` object
        """
        print("Write '%s' file with RegisterMapWriter:" % path)
        print("  Prepare data ... ", end='')
        data = {
            'config': rmap.config.as_dict(),
            'regmap': list(rmap.as_dict().values())
        }
        print("OK")
        self._save_file(path, data)


class ConfigurationWriter(_DictWriter):
    """Write configuration to a file.

    Examples:

        Create JSON file based on a :class:`Configuration` object:

        >>> from corsair import Configuration, ConfigurationWriter
        >>> writer = ConfigurationWriter()
        >>> config = Configuration()
        >>> writer('/tmp/config.json', config)
        Write '/tmp/config.json' file with ConfigurationWriter:
          Prepare data ... OK
          Save data to file ... OK
    """
    def __call__(self, path, config):
        """Write output file.

        Args:
            path : path to file
            rmap : :class:`Configuration` object
        """
        print("Write '%s' file with ConfigurationWriter:" % path)
        print("  Prepare data ... ", end='')
        data = config.as_dict()
        print("OK")
        self._save_file(path, data)


class _Jinja2Writer():
    """Basic class for rendering Jinja2 templates."""

    def _render_to_file(self, template, vars, path):
        """Render text with Jinja2 and save it to file

        Args:
            template : path to Jinja2 template
            vars : dictionary with template variables
            path : path to output file
        """
        print("  Load template ... ", end='')
        templates_path = str(Path(__file__).parent / 'templates')
        j2_env = jinja2.Environment(loader=jinja2.FileSystemLoader(searchpath=templates_path),
                                    trim_blocks=True, lstrip_blocks=True)
        j2_template = j2_env.get_template(template)
        print("OK")

        print("  Render text ... ", end='')
        rendered_text = j2_template.render(vars)
        print("OK")

        print("  Save data to file ... ", end='')
        with open(path, "w") as f:
            f.write(rendered_text)
        print("OK")


class LbBridgeWriter(_Jinja2Writer):
    """Create HDL file with bridge to Local Bus.

    Examples:

        Create Verilog file with APB to Local Bus bridge:

        >>> from corsair import Configuration, LbBridgeWriter
        >>> config = Configuration()
        >>> config['lb_bridge']['type'].value = 'apb'
        >>> writer = LbBridgeWriter()
        >>> writer('/tmp/lb_bridge.v', config)
        Write '/tmp/lb_bridge.v' file with LbBridgeWriter:
          Prepare data ... OK
          Load template ... OK
          Render text ... OK
          Save data to file ... OK
    """
    def __call__(self, path, config):
        """Create bridge to Local Bus in Verilog."""

        if config['lb_bridge']['type'].value == 'axil':
            j2_template = 'axil2lb_verilog.j2'
        elif config['lb_bridge']['type'].value == 'apb':
            j2_template = 'apb2lb_verilog.j2'
        elif config['lb_bridge']['type'].value == 'amm':
            j2_template = 'amm2lb_verilog.j2'
        else:
            print("Local Bus is selected for the CSR interface. Bridge will not be generated.")
            return

        print("Write '%s' file with LbBridgeWriter:" % path)
        print("  Prepare data ... ", end='')

        j2_vars = {}

        j2_vars['corsair_ver'] = __version__
        j2_vars['module_name'] = Path(path).stem
        j2_vars['addr_width'] = config['address_width'].value
        j2_vars['data_width'] = config['data_width'].value

        print("OK")

        self._render_to_file(j2_template, j2_vars, path)


class HdlWriter(_Jinja2Writer):
    """Create HDL file with register map.

    Examples:
        >>> from corsair import RegisterMap, HdlWriter
        >>> rmap = RegisterMap()
        >>> writer = HdlWriter()
        >>> writer('/tmp/regs.v', rmap)
        Write '/tmp/regs.v' file with HdlWriter:
          Prepare data ... OK
          Load template ... OK
          Render text ... OK
          Save data to file ... OK
    """
    def __call__(self, path, rmap):
        """Create register map in Verilog."""
        j2_template = 'regmap_verilog.j2'

        print("Write '%s' file with HdlWriter:" % path)
        print("  Prepare data ... ", end='')

        j2_vars = {}

        j2_vars['corsair_ver'] = __version__
        if not rmap.config['name'].value:
            rmap.config['name'].value = Path(path).stem
        j2_vars['rmap'] = rmap
        j2_vars['config'] = rmap.config

        print("OK")

        self._render_to_file(j2_template, j2_vars, path)


class DocsWriter(_Jinja2Writer):
    """Create documentation for a register map.

    Examples:
        >>> from corsair import RegisterMap, DocsWriter
        >>> rmap = RegisterMap()
        >>> writer = DocsWriter()
        >>> writer('/tmp/regs.md', rmap)
        Write '/tmp/regs.md' file with DocsWriter:
          Prepare data ... OK
          Load template ... OK
          Render text ... OK
          Save data to file ... OK
          Draw images ... OK
    """
    def __call__(self, path, rmap):
        """Create documentation for a register map in Markdown."""
        j2_template = 'regmap_md.j2'

        print("Write '%s' file with DocsWriter:" % path)
        print("  Prepare data ... ", end='')

        j2_vars = {}

        j2_vars['corsair_ver'] = __version__
        if not rmap.config['name'].value:
            rmap.config['name'].value = Path(path).stem
        j2_vars['rmap'] = rmap
        j2_vars['config'] = rmap.config

        print("OK")

        self._render_to_file(j2_template, j2_vars, path)

        print("  Draw images ... ", end='')
        self._draw_regs(Path(path).parent, rmap)
        print("OK")

    def _draw_regs(self, outdir, rmap):
        imgdir = outdir / ('%s_img' % rmap.config['name'].value)
        imgdir.mkdir(exist_ok=True)

        bits = rmap.config['data_width'].value
        lanes = bits // 16 if bits > 16 else 1
        for reg in rmap:
            reg_wd = {"reg": [],
                      "options": {"bits": bits, "lanes": lanes}}
            bit_pos = -1
            for bf in reg:
                if bit_pos == -1 and bf.lsb > 0:
                    reg_wd["reg"].append({"bits": bf.lsb})
                elif bf.lsb - bit_pos > 1:
                    reg_wd["reg"].append({"bits": bf.lsb - bit_pos - 1})
                reg_wd["reg"].append({"name": bf.name, "attr": bf.access, "bits": bf.width})
                bit_pos = bf.msb
            if (bits - 1) > bit_pos:
                reg_wd["reg"].append({"bits": bits - bit_pos - 1})
            wavedrom.render(json.dumps(reg_wd)).saveas(str(imgdir / ("%s.svg" % reg.name.lower())))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Configuration module.
"""

from . import utils


class Parameter():
    """Generic parameter.

    Examples:

        Create parameter with check that value never exceed 42:

        >>> print(Parameter('param_name', 42, lambda val: val <= 42))
        param_name: 42

    Attributes:
        name: Name of the parameter.
        validator: Lambda function to perform value validation.
            Returns True if value check is OK, and False otherwise.
    """
    def __init__(self, name, value=None, validator=lambda val: True):
        self._value = None

        self.name = name
        self.validator = validator
        self.value = value

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" % (repr(self), repr(other)))
        else:
            return (
                self.as_dict() == other.as_dict() and
                self.validator == other.validator
            )

    def __ne__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" % (repr(self), repr(other)))
        else:
            return not self.__eq__(other)

    def __repr__(self):
        return 'Parameter(%s, %s)' % (repr(self.name), repr(self.value))

    def __str__(self):
        return self.as_str()

    def as_str(self, indent=''):
        """Returns indented parameter's string with name and value."""
        return indent + '%s: %s' % (self.name, utils.try_int_to_str(self._value))

    def as_dict(self):
        """Returns dictionary with parameters's name and value."""
        return {'name': self.name, 'value': self.value}

    @property
    def value(self):
        """Value of the parameter.

        Getter:
            Get current value of the parameter.

        Setter:
            Set new value of the parameter.

            Raises:
                ValueError: An error occured if new value fails the check.
        """
        return self._value

    @value.setter
    def value(self, new_value):
        # store all 0x-like hexademical strings as integers if possible
        self._value = utils.try_hex_to_dec(new_value)
        # run validator for the new value
        self._validate()

    def _validate(self):
        """Check parameter value correctness."""
        if self.validator(self.value) is False:
            raise ValueError('"%s" parameter with "%s" value failed the check!' % (self.name,
                                                                                   self.value))


class ParameterGroup():
    """Group of parameters or other groups.

    Examples:

        Create group and add parameters:

        >>> pg = ParameterGroup('group_a')
        >>> pg.add_params([Parameter('p1', 42), Parameter('p2', 'abc')])
        >>> print(pg)
        group_a:
          p1: 42
          p2: abc

        Other way to do the same:

        >>> pg = ParameterGroup('group_a')
        >>> pg.values = {'p1':42, 'p2':'abc'}
        >>> print(pg)
        group_a:
          p1: 42
          p2: abc

        Change parameter value:

        >>> pg = ParameterGroup('group_a')
        >>> pg.add_params(Parameter('p1', 42))
        >>> pg['p1'].value = 777
        >>> print(pg['p1'].value)
        777

    Attributes:
        name: Name of the group.
    """
    def __init__(self, name):
        self.name = name
        self._params = {}

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" % (repr(self), repr(other)))
        else:
            return self.as_dict() == other.as_dict()

    def __ne__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" % (repr(self), repr(other)))
        else:
            return not self.__eq__(other)

    def __repr__(self):
        return 'ParameterGroup(%s)' % repr(self.name)

    def __str__(self):
        return self.as_str()

    def __getitem__(self, key):
        try:
            return self._params[key]
        except KeyError:
            raise KeyError("Parameter/Group with a name '%s' doesn't exist!" % key)

    def __setitem__(self, key, value):
        raise KeyError("Not able to set '%s' item directly in '%s' group!"
                       " Try use add_params() method." % (key, self.name))

    def as_str(self, indent=''):
        """Returns indented string with group members and their values"""
        new_indent = indent + '  '
        params = [self[name].as_str(new_indent) for name in self.names]
        params_str = '\n'.join(params) if params else new_indent + 'empty'
        return indent + '%s:\n' % self.name + params_str

    def as_dict(self):
        """Returns dictionary with group's key attributes."""
        return self.values

    @property
    def names(self):
        """Names of all parameters/groups contained."""
        return self._params.keys()

    @property
    def params(self):
        """List with parameters/groups objects contained."""
        return [param for param in self._params.items()]

    def add_params(self, new_params):
        """Add parameters.

        Args:
            new_params : list with parameters/groups or single object
        """
        # hack to handle single elements
        new_params = utils.listify(new_params)

        # add params to dict one by one
        for p in new_params:
            if p.name in self._params:
                raise KeyError("Item with name '%s' is already present in '%s' group!" % (p.name, self.name))
            self._params[p.name] = p

    @property
    def values(self):
        """Dictionary with values of the parameters/groups.

        Getter:
            Returns dictionary with pairs {name:value} of all parameter/groups of the current group.

        Setter:
            Sets only parameters/groups specified in the input dictionary,
            but if parameter/group does not exist, it will create it.
        """
        values = {}
        for name in self.names:
            if isinstance(self[name], ParameterGroup):
                values[name] = self[name].values
            else:
                values[name] = self[name].value
        return values

    @values.setter
    def values(self, new_values):
        for name in new_values.keys():
            try:
                # if parameter/group exists
                if isinstance(self[name], ParameterGroup):
                    self[name].values = new_values[name]
                else:
                    self[name].value = new_values[name]
            except KeyError:
                # else if parameter/group does not exists
                if isinstance(new_values[name], dict):
                    self.add_params(ParameterGroup(name=name))
                    self[name].values = new_values[name]
                else:
                    self.add_params(Parameter(name=name, value=new_values[name]))


class Configuration(ParameterGroup):
    """Collection of global parameters."""
    def __init__(self):
        super().__init__('configuration')
        self._init_default_params()

    def __repr__(self):
        return 'Configuration()'

    def _init_default_params(self):
        """Initalize all default params"""
        # group regmap
        self.add_params(ParameterGroup('regmap'))
        self['regmap'].add_params([
            Parameter(name='read_filler', value=0x0),
            Parameter(name='address_increment_mode', value='none',
                      validator=lambda val: val in ['none', 'data_width', 'custom']),
            Parameter(name='address_increment_value', value=4, validator=lambda val: val >= 1),
            Parameter(name='address_alignment_mode', value='data_width',
                      validator=lambda val: val in ['none', 'data_width', 'custom']),
            Parameter(name='address_alignment_value', value=4, validator=lambda val: val >= 1)
        ])

        # group lb_bridge
        self.add_params(ParameterGroup('lb_bridge'))
        lb_bridge_type_allowed = ['amm', 'apb', 'axil', 'none']
        self['lb_bridge'].add_params(
            Parameter(name='type', value='none', validator=lambda val: val in lb_bridge_type_allowed)
        )

        # group docs
        self.add_params(ParameterGroup('docs'))

        # common params
        data_width_allowed = {
            'amm': [8, 16, 32, 64, 128, 256, 512, 1024],
            'apb': [8, 16, 32],
            'axil': [32, 64],
            'none': [8, 16, 32, 64, 128, 256, 512, 1024]
        }
        self.add_params(Parameter(
            name='data_width',
            value=32,
            validator=lambda val: val in data_width_allowed[self['lb_bridge']['type'].value])
        )

        addr_width_allowed = {
            'amm': range(1, 65),
            'apb': range(1, 33),
            'axil': [32, 64],
            'none': range(1, 65)
        }
        self.add_params(Parameter(
            name='address_width',
            value=32,
            validator=lambda val: val in addr_width_allowed[self['lb_bridge']['type'].value])
        )

        self.add_params([
            Parameter(name='name', value=''),
            Parameter(name='version', value='1.0')]
        )

        reg_rst_allowlist = ['sync_pos', 'sync_neg', 'async_pos', 'async_neg', 'init_only']
        self.add_params(Parameter(name='register_reset', value='sync_pos',
                                  validator=lambda val: val in reg_rst_allowlist))



import setuptools
from pkg_resources import parse_version

VERSION = "0.2.0"


# Based on https://github.com/tulip-control/dd/blob/885a716a56e82bfee54b0178d0ce38298b85eb6a/setup.py#L68
def git_version(version):
    """Return version with local version identifier."""
    import git
    repo = git.Repo('.git')
    repo.git.status()
    # assert versions are increasing
    latest_tag = repo.git.describe(
        match='v[0-9]*', tags=True, abbrev=0)
    assert parse_version(latest_tag) <= parse_version(version), (
        latest_tag, version)
    sha = repo.head.commit.hexsha[:8]
    if repo.is_dirty():
        return '{v}.dev0+{sha}.dirty'.format(
            v=version, sha=sha)
    # commit is clean
    # is it release of `version` ?
    try:
        tag = repo.git.describe(
            match='v[0-9]*', exact_match=True,
            tags=True, dirty=True)
    except git.GitCommandError:
        return '{v}.dev0+{sha}'.format(
            v=version, sha=sha)
    assert tag == 'v' + version, (tag, version)
    return version


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# Get version
VERSION_FILE = 'corsair/_version.py'
try:
    version = git_version(VERSION)
except AssertionError:
    print('No git info: Assume release.')
    version = VERSION
with open(VERSION_FILE, 'w') as f:
    f.write("version = '%s'\n" % version)

# Install package
setuptools.setup(
    name="corsair",
    version=version,
    author="esynr3z",
    author_email="esynr3z@gmail.com",
    description="Control and Status Register map generator for FPGA/ASIC projects",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/esynr3z/corsair",
    project_urls={
        'Documentation': 'https://corsair.readthedocs.io'
    },
    packages=setuptools.find_packages(exclude='tests'),
    package_data={'corsair': ['templates/*.j2']},
    entry_points={
        'console_scripts': [
            'corsair = corsair.__main__:main',
        ],
    },
    install_requires=[
        'pyyaml',
        'jinja2',
        'wavedrom',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)

# Clear version info
with open(VERSION_FILE, 'w') as f:
    f.write("")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Register map.
"""

from . import utils
from .config import Configuration


class BitField():
    """Bit field.

    Examples:

        Create a bit field:

        >>> print(BitField('bf_a', 'bf_a description', lsb=0, width=8))
        bf_a: bf_a description
          initial = 0
          width = 8
          lsb = 0
          access = rw
          access_flags = False
          modifiers = []

    Attributes:
        name: Name of the bit field.
        description: Description of the bit field.
    """
    def __init__(self, name, description='', initial=0, width=1,
                 lsb=0, access='rw', access_flags=False, modifiers=[]):
        self.name = name
        self.description = description
        self.initial = initial
        self.width = width
        self.lsb = lsb
        self.access = access
        self.access_flags = access_flags
        self.modifiers = modifiers

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" % (repr(self), repr(other)))
        else:
            return self.as_dict() == other.as_dict()

    def __ne__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" % (repr(self), repr(other)))
        else:
            return not self.__eq__(other)

    def __repr__(self):
        return 'BitField(%s)' % repr(self.name)

    def __str__(self):
        return self.as_str()

    def as_str(self, indent=''):
        """Returns indented string with the bit field information."""
        inner_indent = indent + '  '
        bf_str = indent + '%s: %s\n' % (self.name, self.description)
        bf_str += inner_indent + 'initial = %s\n' % utils.try_int_to_str(self.initial)
        bf_str += inner_indent + 'width = %s\n' % self.width
        bf_str += inner_indent + 'lsb = %s\n' % self.lsb
        bf_str += inner_indent + 'access = %s\n' % self.access
        bf_str += inner_indent + 'access_flags = %s\n' % self.access_flags
        bf_str += inner_indent + 'modifiers = %s' % self.modifiers
        return bf_str

    def as_dict(self):
        """Returns dictionary with bit field's key attributes."""
        return {
            'name': self.name,
            'description': self.description,
            'initial': self.initial,
            'width': self.width,
            'lsb': self.lsb,
            'access': self.access,
            'access_flags': self.access_flags,
            'modifiers': self.modifiers
        }

    @property
    def initial(self):
        """Initial value for the field. Only non-negative integers are allowed."""
        return self._initial

    @initial.setter
    def initial(self, value):
        value = utils.try_hex_to_dec(value)
        err_msg = ("Initial value '%s' for '%s' is wrong!"
                   " Only non-negative integers are allowed." % (value, self.name))
        if utils.is_non_neg_int(value, err_msg):
            self._initial = value

    @property
    def width(self):
        """Bit width of the field. Only positive integers are allowed."""
        return self._width

    @width.setter
    def width(self, value):
        value = utils.try_hex_to_dec(value)
        err_msg = ("Width value '%s' for '%s' is wrong!"
                   " Only positive integers are allowed." % (value, self.name))
        if utils.is_pos_int(value, err_msg):
            self._width = value

    @property
    def lsb(self):
        """Position of less significant bit (LSB) of the field. Only non-negative integers are allowed."""
        return self._lsb

    @lsb.setter
    def lsb(self, value):
        value = utils.try_hex_to_dec(value)
        err_msg = ("LSB value '%s' for '%s' is wrong!"
                   " Only non-negative integers are allowed." % (value, self.name))
        if utils.is_non_neg_int(value, err_msg):
            self._lsb = value

    @property
    def msb(self):
        """Position of most significant bit (MSB) of the field."""
        return self.lsb + self.width - 1

    @property
    def byte_strobes(self):
        """Dictionary with LSB and MSB values for every byte in write data bus."""
        strb = {}
        first = self.lsb // 8
        last = self.msb // 8
        for i in range(first, last + 1):
            # per every byte strobe
            wdata_lsb = self.lsb if i == first else i * 8
            wdata_msb = (i + 1) * 8 - 1 if ((i + 1) * 8 - 1 - self.msb) < 0 else self.msb
            bf_lsb = wdata_lsb - self.lsb
            bf_msb = wdata_msb - self.lsb
            strb[i] = {'bf_lsb': bf_lsb, 'bf_msb': bf_msb,
                       'wdata_lsb': wdata_lsb, 'wdata_msb': wdata_msb}
        return strb

    @property
    def access(self):
        """Bit field access mode."""
        return self._access

    @access.setter
    def access(self, value):
        if value not in ['rw', 'ro', 'wo']:
            raise ValueError("Unknown access mode '%s' for '%s' field!" % (value, self.name))
        else:
            self._access = value

    @property
    def access_flags(self):
        """Enable access flags generation."""
        return self._access_flags

    @access_flags.setter
    def access_flags(self, value):
        if isinstance(value, bool):
            self._access_flags = value
        else:
            raise ValueError("Access flags attribute has to be 'bool', "
                             "but '%s' provided for '%s' field!" % (type(value), self.name))

    @property
    def modifiers(self):
        """List of an access modifiers."""
        return self._modifiers

    @modifiers.setter
    def modifiers(self, value):
        # hack to handle single elements
        if type(value) is not list:
            value = [value]

        # check if all options are allowed
        allowlist = [
            'sc',
            'w1tc',
            'w1ts',
            'w1tt',
            'rtc',
            'const',
            'hwu',
        ]
        for v in value:
            if v not in allowlist:
                raise ValueError("Unknown access mode '%s' for '%s' field!" % (v, self.name))

        # check if options combination is allowed
        allowlist_comb = {
            'rw': [
                [],
                ['hwu'],
                ['hwu', 'w1tc'],
                ['hwu', 'w1ts'],
                ['hwu', 'w1tt'],
            ],
            'wo': [
                [],
                ['sc'],
            ],
            'ro': [
                [],
                ['const'],
                ['hwu'],
                ['hwu', 'rtc'],
            ],
        }
        if value not in allowlist_comb[self.access]:
            raise ValueError("Unknown access modifiers combination '%s' for '%s' field!" % (value, self.name))

        self._modifiers = value

    @property
    def bits(self):
        """Returns list with all bits positions used by a bit field"""
        return list(range(self.lsb, self.msb + 1))

    def is_vector(self):
        """Returns True if bit fields's width > 1."""
        return True if self.width > 1 else False


class Register():
    """Control and status register.

    Examples:

        Create a register:

        >>> reg = Register('reg_a', 'Register A description', address=8)
        >>> reg.add_bfields([
        ...     BitField('bf_a', 'Bit field A', lsb=0),
        ...     BitField('bf_b', 'Bit field B', lsb=1)
        ... ])
        >>> print(reg)
        (0x8) reg_a: Register A description
          bf_a: Bit field A
            initial = 0
            width = 1
            lsb = 0
            access = rw
            access_flags = False
            modifiers = []
          bf_b: Bit field B
            initial = 0
            width = 1
            lsb = 1
            access = rw
            access_flags = False
            modifiers = []

        Access bit field via name or index:

        >>> reg = Register('reg_a', 'Register A description', address=8)
        >>> reg.add_bfields([
        ...     BitField('bf_a', 'Bit field A', lsb=0),
        ...     BitField('bf_b', 'Bit field B', lsb=1)
        ... ])
        >>> print(reg['bf_b'].name, reg[1].name)
        bf_b bf_b

        Iterate through bit fields:

        >>> reg = Register('reg_a', 'Register A description', address=8)
        >>> reg.add_bfields([
        ...     BitField('bf_a', 'Bit field A', lsb=0),
        ...     BitField('bf_b', 'Bit field B', lsb=1)
        ... ])
        >>> for bf in reg:
        ...     print(bf.name)
        ...
        bf_a
        bf_b
    """
    def __init__(self, name='', description='', address=None):
        self._bfields = []

        self.name = name
        self.description = description
        self.address = address

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" % (repr(self), repr(other)))
        else:
            return self.as_dict() == other.as_dict()

    def __ne__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" % (repr(self), repr(other)))
        else:
            return not self.__eq__(other)

    def __repr__(self):
        return 'Register(%s, %s, %s)' % (repr(self.name), repr(self.description), repr(self.address))

    def __str__(self):
        return self.as_str()

    def as_str(self, indent=''):
        """Returns indented string with the register information."""
        inner_indent = indent + '  '
        bfields = [bf.as_str(inner_indent) for bf in self.bfields]
        bfields_str = '\n'.join(bfields) if bfields else inner_indent + 'empty'
        return indent + '(0x%x) %s: %s\n' % (self.address, self.name, self.description) + bfields_str

    def as_dict(self):
        """Returns dictionary with register's key attributes."""
        return {
            'name': self.name,
            'description': self.description,
            'address': self.address,
            'bfields': [bf.as_dict() for bf in self.bfields]
        }

    def __len__(self):
        """Number of register's bit fields."""
        return len(self._bfields)

    def __iter__(self):
        """Bit fields iterator."""
        return iter(self._bfields)

    def __getitem__(self, key):
        """Get bit field by name or index.

        Raises:
            KeyError: An error occured if bit field does not exists.
        """
        try:
            if isinstance(key, str):
                return next(bf for bf in self if bf.name == key)
            else:
                return self._bfields[key]
        except (StopIteration, TypeError, KeyError, IndexError):
            raise KeyError("There is no bit field with a name/index '%s' in '%s' register!" % (key, self.name))

    def __setitem__(self, key, value):
        """Set bit field by key"""
        raise KeyError("Not able to set '%s' bit field directly in '%s' register!"
                       " Try use add_bfields() method." % (key, self.name))

    @property
    def name(self):
        """Register name. If name is not set, use name of the first field."""
        if not self._name:
            if len(self) == 0:
                raise ValueError("Register has no name and no fields!")
            elif len(self) > 1:
                raise ValueError("Register has %d fields but has no name!" % len(self))
            else:
                return self[0].name
        else:
            return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def address(self):
        """Register address. Only non-negative integers are allowed."""
        return self._address

    @address.setter
    def address(self, value):
        value = utils.try_hex_to_dec(value)
        err_msg = ("Address value '%s' for '%s' is wrong!"
                   " Only non-negative integers are allowed." % (value, self._name))
        if value is None:
            self._address = None
        elif utils.is_non_neg_int(value, err_msg):
            self._address = value

    @property
    def names(self):
        """Return all bit field names"""
        return [bf.name for bf in self]

    @property
    def description(self):
        """Register description.If description is not set, use description of the first field."""
        if not self._description and len(self) == 1:
            return self[0].description
        else:
            return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def bfields(self):
        """Returns list with bit field objects."""
        return self._bfields

    def add_bfields(self, new_bfields):
        """Add bit field or list of bit feilds.

        Bit fields are automatically sorted and stored in the ascending order of msb attributes.
        """
        # hack to handle single elements
        new_bfields = utils.listify(new_bfields)

        # add bit fields to list one by one
        for bf in new_bfields:
            # check existance
            if bf.name in self.names:
                raise ValueError("Bit field with name '%s' is already present in '%s' register!" % (bf.name, self.name))
            # check fields overlapping
            overlaps = [set(bf.bits).intersection(set(old_bf.bits)) for old_bf in self._bfields]
            overlaps_names = [self.names[i] for i, ovl in enumerate(overlaps) if ovl]
            if self.bfields and overlaps_names:
                raise ValueError("Position of a bit field '%s'"
                                 " conflicts with other bit field(s): %s!" % (bf.name, repr(overlaps_names)))
            # if we here - all is ok and bit field can be added
            try:
                # find position to insert bit field and not to break ascending order of bit field msb positions
                bf_idx = next(i for i, old_bf in enumerate(self._bfields) if old_bf.msb > bf.msb)
                self._bfields.insert(bf_idx, bf)
            except StopIteration:
                # when bit field list is empty or all bit field msb positions are less than the current one
                self._bfields.append(bf)

    @property
    def initial(self):
        """Initial value of the refister after reset."""
        init = 0
        for bf in self:
            init |= bf.initial << bf.lsb
        return init


class RegisterMap():
    """CSR map.

    Examples:

        Create a register map:

        >>> rmap = RegisterMap()
        >>> rmap.config['name'].value = 'regs'
        >>> rmap.add_regs([
        ...     Register('reg_a', 'Register A', address=0),
        ...     Register('reg_b', 'Register B', address=4)
        ... ])
        >>> print(rmap)
        regs: v1.0
          (0x0) reg_a: Register A
            empty
          (0x4) reg_b: Register B
            empty

        Access register via name or index:

        >>> rmap = RegisterMap()
        >>> rmap.add_regs([
        ...     Register('reg_a', 'Register A', address=0),
        ...     Register('reg_b', 'Register B', address=4)
        ... ])
        >>> print(rmap['reg_a'].name, rmap[0].name)
        reg_a reg_a

        Iterate through registers:

        >>> rmap = RegisterMap()
        >>> rmap.add_regs([
        ...     Register('reg_a', 'Register A', address=0),
        ...     Register('reg_b', 'Register B', address=4)
        ... ])
        >>> for reg in rmap:
        ...     print(reg.name)
        ...
        reg_a
        reg_b

    """
    def __init__(self, config=Configuration()):
        self.config = config
        self.name = self.config['name'].value
        self.version = self.config['version'].value
        self._regs = []

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" % (repr(self), repr(other)))
        else:
            return self.as_dict() == other.as_dict()

    def __ne__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" % (repr(self), repr(other)))
        else:
            return not self.__eq__(other)

    def __repr__(self):
        return 'RegisterMap(%s)' % (repr(self.config))

    def __str__(self):
        return self.as_str()

    def as_str(self, indent=''):
        """Returns indented string with the register map information."""
        inner_indent = indent + '  '
        regs = [reg.as_str(inner_indent) for reg in self.regs]
        regs_str = '\n'.join(regs) if regs else inner_indent + 'empty'
        return indent + '%s: v%s\n' % (self.config['name'].value, self.config['version'].value) + regs_str

    def as_dict(self):
        """Returns register map as a dictionary."""
        return {reg.name: reg.as_dict() for reg in self.regs}

    def __len__(self):
        """Number of registers."""
        return len(self._regs)

    def __iter__(self):
        """Registers iterator."""
        return iter(self._regs)

    def __getitem__(self, key):
        """Get register by name or index.

        Raises:
            KeyError: An error occured if register does not exists.
        """
        try:
            if isinstance(key, str):
                return next(reg for reg in self if reg.name == key)
            else:
                return self._regs[key]
        except (StopIteration, TypeError, KeyError, IndexError):
            raise KeyError("There is no register with a name/index '%s'!" % (key))

    def __setitem__(self, key, value):
        """Set register by key"""
        raise KeyError("Not able to set '%s' register directly!"
                       " Try use add_regs() method." % (key))

    @property
    def names(self):
        """Return all register names."""
        return [reg.name for reg in self]

    def _addr_apply(self, reg):
        """Apply auto-calculated address for a register with no address."""
        # some error checks
        if len(self) == 0:
            raise ValueError("Register '%s' with no address is not allowed"
                             " to be the first register in a map!" % (reg.name))
        if self.config['regmap']['address_increment_mode'].value == 'none':
            raise ValueError("Register '%s' with no address is not allowed"
                             " when address auto increment is disabled!" % (reg.name))

        prev_addr = self.regs[-1].address

        if self.config['regmap']['address_increment_mode'].value == 'data_width':
            addr_step = self.config['data_width'].value // 8
        else:
            addr_step = self.config['regmap']['address_increment_value'].value

        reg.address = prev_addr + addr_step

    def _addr_check(self, reg):
        """Check address alignment."""
        if self.config['regmap']['address_alignment_mode'].value == 'data_width':
            align_val = self.config['data_width'].value // 8
        elif self.config['regmap']['address_alignment_mode'].value == 'custom':
            align_val = self.config['regmap']['address_alignment_value'].value
        else:
            align_val = 1

        if (reg.address % align_val) != 0:
            raise ValueError("Register '%s' with address '%d' is not %d bytes alligned!" %
                             (reg.name, reg.address, align_val))

    @property
    def regs(self):
        """Returns list with register objects."""
        return self._regs

    def add_regs(self, new_regs):
        """Add register or list of registers.

        Register are automatically sorted and stored in the ascending order of addresses.
        """
        # hack to handle single elements
        new_regs = utils.listify(new_regs)

        # add registers to list one by one
        for reg in new_regs:
            # check existance
            if reg.name in self.names:
                raise ValueError("Register with name '%s' is already present!" % (reg.name))
            # check bit field conflicts with data width
            for bf in reg:
                if bf.msb >= self.config['data_width'].value:
                    raise ValueError("Register '%s' has field '%s' (msb=%d) "
                                     "that exceeds interface data width %d!" %
                                     (reg.name, bf.name, bf.msb, self.config['data_width'].value))
            # aplly calculated address if register address is empty
            if reg.address is None:
                self._addr_apply(reg)
            # check address alignment
            self._addr_check(reg)
            # check address conflicts
            addresses = [reg.address for reg in self]
            if reg.address in addresses:
                conflict_reg = self[addresses.index(reg.address)].name
                raise ValueError("Register '%s' with address '%d'"
                                 " conflicts with register '%s' with the same address!" %
                                 (reg.name, reg.address, conflict_reg))
            # if we here - all is ok and register can be added
            try:
                # find position to insert register and not to break ascending order of addresses
                reg_idx = next(i for i, r in enumerate(self._regs) if r.address > reg.address)
                self._regs.insert(reg_idx, reg)
            except StopIteration:
                # when registers list is empty or all addresses are less than the current one
                self._regs.append(reg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Run Corsair from command line with arguments.
"""

import sys
import os
import argparse
from pathlib import Path
import corsair
from . import utils

__all__ = ['main']


class ArgumentParser(argparse.ArgumentParser):
    """Inherit ArgumentParser to override the behaviour of error method."""
    def error(self, message):
        self.print_help(sys.stderr)
        self.exit(2, '\n%s: error: %s\n' % (self.prog, message))


def parse_arguments():
    """Parse and validate arguments."""

    parser = ArgumentParser(prog=corsair.__title__,
                            description=corsair.__description__)
    parser.add_argument('-v', '--version',
                        action='version',
                        version='%(prog)s v' + corsair.__version__)
    parser.add_argument('-r', '--regmap',
                        metavar='<file>',
                        dest='regmap',
                        help='read register map from input <file>')
    parser.add_argument('-c', '--config',
                        metavar='<file>',
                        dest='config',
                        help='read configuration from <file>')
    parser.add_argument('--template-regmap',
                        metavar='<file>',
                        dest='template_regmap',
                        help='create register map template <file>')
    parser.add_argument('--template-config',
                        metavar='<file>',
                        dest='template_config',
                        help='create configuration template <file>')
    parser.add_argument('--dump-regmap',
                        metavar='<file>',
                        dest='dump_regmap',
                        help='dump register map to <file>')
    parser.add_argument('--dump-config',
                        metavar='<file>',
                        dest='dump_config',
                        help='dump configuration to <file>')
    parser.add_argument('--output-dir',
                        metavar='<dir>',
                        dest='outdir',
                        help='save all generated artifacts to <dir>')
    parser.add_argument('--hdl',
                        dest='hdl',
                        action='store_true',
                        help='generate HDL module with register map')
    parser.add_argument('--lb-bridge',
                        dest='lb_bridge',
                        action='store_true',
                        help='generate HDL module with bridge to LocalBus')
    parser.add_argument('--docs',
                        dest='docs',
                        action='store_true',
                        help='generate docs for register map')

    # check if no arguments provided
    if len(sys.argv) == 1:
        parser.parse_args(['--help'])
        # argparse will raise SystemExit here

    # get arguments namespace
    args = parser.parse_args()

    # check conflicts
    if not args.regmap and (args.hdl or args.docs):
        parser.error("Not able to proceed without -r/--regmap argument!")

    if args.lb_bridge and not (args.regmap or args.config):
        parser.error("Not able to proceed without -r/--regmap or -c/--config argument!")

    if not args.regmap and args.dump_regmap:
        parser.error("Not able to proceed without -r/--regmap argument!")

    if args.dump_config and not (args.regmap or args.config):
        parser.error("Not able to proceed without -r/--regmap or -c/--config argument!")

    return args


def main():
    """Program main."""
    # parse arguments
    args = parse_arguments()

    # create templates
    if args.template_regmap:
        regs = [corsair.Register('spam', 'Register spam', 0),
                corsair.Register('eggs', 'Register eggs', 4)]
        regs[0].add_bfields([
            corsair.BitField('foo', 'Bit field foo', lsb=0, width=7, access='rw', initial=42),
            corsair.BitField('bar', 'Bit field bar', lsb=24, width=1, access='wo', modifiers=['sc'])
        ])
        regs[1].add_bfields(corsair.BitField('baz', 'Bit field baz', lsb=16, width=16, access='ro'))
        rmap = corsair.RegisterMap()
        rmap.add_regs(regs)
        corsair.RegisterMapWriter()(args.template_regmap, rmap)

    if args.template_config:
        config = corsair.Configuration()
        corsair.ConfigurationWriter()(args.template_config, config)

    # parse input files
    if args.config:
        config = corsair.ConfigurationReader()(args.config)
    else:
        config = corsair.Configuration()

    if args.regmap:
        rmap = corsair.RegisterMapReader()(args.regmap, config)
    else:
        rmap = corsair.RegisterMap(config)

    # dump files
    if args.dump_regmap:
        corsair.RegisterMapWriter()(args.dump_regmap, rmap)

    if args.dump_config:
        corsair.ConfigurationWriter()(args.dump_config, rmap.config)

    # prepare for artifacts generation
    if args.regmap or args.config:
        # output directory
        if args.outdir:
            outdir = Path(args.outdir)
        elif args.regmap:
            outdir = Path(args.regmap).parent
        elif args.config:
            outdir = Path(args.config).parent
        # output file base name
        if config['name'].value:
            outname = config['name'].value
        elif args.regmap:
            outname = Path(args.regmap).stem
            config['name'].value = outname
        elif args.config:
            outname = Path(args.config).stem
            config['name'].value = outname

    # create register map HDL
    if args.hdl:
        hdl_name = '%s.v' % outname
        hdl_path = str(outdir / hdl_name)
        corsair.HdlWriter()(hdl_path, rmap)

    # create bridge to LocalBus HDL
    if args.lb_bridge:
        if config['lb_bridge']['type'].value == 'none':
            print("Local Bus bridge type is 'none' - no need to generate any file.")
        else:
            lb_bridge_name = '%s2lb_%s.v' % (config['lb_bridge']['type'].value, outname)
            lb_bridge_path = str(outdir / lb_bridge_name)
            corsair.LbBridgeWriter()(lb_bridge_path, config)

    # create docs
    if args.docs:
        doc_name = '%s.md' % outname
        doc_path = str(outdir / doc_name)
        corsair.DocsWriter()(doc_path, rmap)

    sys.exit(0)


if __name__ == '__main__':
    main()
